@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2020 by Farsight Security, Inc.
+ * Copyright (c) 2014-2021 by Farsight Security, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -46,15 +46,10 @@
 /* Types. */
 
 #define MAIN_PROGRAM
+#include "asinfo.h"
 #include "defs.h"
 #include "netio.h"
 #include "pdns.h"
-#if WANT_PDNS_DNSDB || WANT_PDNS_DNSDB2
-#include "pdns_dnsdb.h"
-#endif
-#if WANT_PDNS_CIRCL
-#include "pdns_circl.h"
-#endif
 #include "sort.h"
 #include "time.h"
 #include "globals.h"
@@ -65,7 +60,6 @@
 /* Forward. */
 
 static void help(void);
-static pdns_system_ct pick_system(const char *);
 static void qdesc_debug(const char *, qdesc_ct);
 static void qparam_debug(const char *, qparam_ct);
 static __attribute__((noreturn)) void usage(const char *, ...);
@@ -73,14 +67,14 @@ static bool parse_long(const char *, long *);
 static const char *qparam_ready(qparam_t);
 static const char *qparam_option(int, const char *, qparam_t);
 static verb_ct find_verb(const char *);
-static void read_configs(void);
+static char *select_config(void);
 static void do_batch(FILE *, qparam_ct);
 static const char *batch_options(const char *, qparam_t, qparam_ct);
 static const char *batch_parse(char *, qdesc_t);
-static char *makepath(mode_e, const char *, const char *,
-		      const char *, const char *);
+static char *makepath(qdesc_ct);
 static query_t query_launcher(qdesc_ct, qparam_ct, writer_t);
-static void launch(query_t, pdns_fence_ct);
+static const char *rrtype_correctness(const char *);
+static void launch_fetch(query_t, const char *, pdns_fence_ct);
 static void ruminate_json(int, qparam_ct);
 static const char *lookup_ok(void);
 static const char *summarize_ok(void);
@@ -99,10 +93,16 @@ static const char * const conf_files[] = {
 const struct verb verbs[] = {
 	/* note: element [0] of this array is the DEFAULT_VERB. */
 	{ "lookup", "/lookup", lookup_ok,
-	  present_text_lookup, present_json, present_csv_lookup },
+	  present_text_lookup,
+	  present_json_lookup,
+	  present_csv_lookup,
+	  present_minimal_lookup },
 	{ "summarize", "/summarize", summarize_ok,
-	  present_text_summarize, present_json, present_csv_summarize },
-	{ NULL, NULL, NULL, NULL, NULL, NULL }
+	  present_text_summarize,
+	  present_json_summarize,
+	  present_csv_summarize,
+	  NULL },
+	{ NULL, NULL, NULL, NULL, NULL, NULL, NULL }
 };
 
 /* Private. */
@@ -116,6 +116,7 @@ int
 main(int argc, char *argv[]) {
 	struct qdesc qd = { .mode = no_mode };
 	struct qparam qp = qparam_empty;
+	char *picked_system = NULL;
 	bool info = false;
 	int json_fd = -1;
 	const char *msg;
@@ -132,12 +133,17 @@ main(int argc, char *argv[]) {
 	value = getenv(env_time_fmt);
 	if (value != NULL && strcasecmp(value, "iso") == 0)
 		iso8601 = true;
+
+	value = getenv(env_config_file);
+	if (value != NULL)
+		config_file = strdup(value);
+
 	pverb = &verbs[DEFAULT_VERB];
 
 	/* process the command line options. */
 	while ((ch = getopt(argc, argv,
-			    "R:r:N:n:i:M:u:p:t:b:k:J:V:"
-			    "dfhIjmqSsUv468" QPARAM_GETOPT))
+			    "D:R:r:N:n:i:M:u:p:t:b:k:J:V:T:0:"
+			    "adfhIjmqSsUv468" QPARAM_GETOPT))
 	       != -1)
 	{
 		switch (ch) {
@@ -149,6 +155,27 @@ main(int argc, char *argv[]) {
 			if ((msg = qparam_option(ch, optarg, &qp)) != NULL)
 				usage(msg);
 			break;
+		case '0': {
+			const char *equal = strchr(optarg, '=');
+			if (equal == NULL)
+				usage("-0 requires 'function='");
+			size_t length = (size_t) (equal - optarg);
+			const char *thing = optarg + length + 1;
+			if (strncmp(optarg, "countoff", length) == 0) {
+				struct counted *c = countoff(thing);
+				countoff_debug("main", thing, c);
+				DESTROY(c);
+			} else {
+				usage("-0 function unrecognized");
+			}
+			my_exit(0);
+		    }
+		case 'a':
+			asinfo_lookup = true;
+			break;
+		case 'D':
+			asinfo_domain = optarg;
+			break;
 		case 'R': {
 			if (qd.mode != no_mode)
 				usage("-r, -n, -i, -N, or -R "
@@ -166,13 +193,13 @@ main(int argc, char *argv[]) {
 				const char *q = strchr(p + 1, '/');
 				if (q != NULL) {
 					qd.bailiwick = strdup(q + 1);
-					qd.rrtype = strndup(p + 1,
-							 (size_t)(q - p - 1));
+					qd.rrtype = strndup(p + 1, (size_t)
+							    (q - p - 1));
 				} else {
 					qd.rrtype = strdup(p + 1);
 				}
 				qd.thing = strndup(optarg,
-						      (size_t)(p - optarg));
+						   (size_t)(p - optarg));
 			} else {
 				qd.thing = strdup(optarg);
 			}
@@ -195,13 +222,13 @@ main(int argc, char *argv[]) {
 				const char *q = strchr(p + 1, '/');
 				if (q != NULL) {
 					qd.bailiwick = strdup(q + 1);
-					qd.rrtype = strndup(p + 1,
-							 (size_t)(q - p - 1));
+					qd.rrtype = strndup(p + 1, (size_t)
+							    (q - p - 1));
 				} else {
 					qd.rrtype = strdup(p + 1);
 				}
 				qd.thing = strndup(optarg,
-						      (size_t)(p - optarg));
+						   (size_t)(p - optarg));
 			} else {
 				qd.thing = strdup(optarg);
 			}
@@ -224,13 +251,13 @@ main(int argc, char *argv[]) {
 				const char *q = strchr(p + 1, '/');
 				if (q != NULL) {
 					qd.bailiwick = strdup(q + 1);
-					qd.rrtype = strndup(p + 1,
-							 (size_t)(q - p - 1));
+					qd.rrtype = strndup(p + 1, (size_t)
+							    (q - p - 1));
 				} else {
 					qd.rrtype = strdup(p + 1);
 				}
 				qd.thing = strndup(optarg,
-						      (size_t)(p - optarg));
+						   (size_t)(p - optarg));
 			} else {
 				qd.thing = strdup(optarg);
 			}
@@ -253,13 +280,13 @@ main(int argc, char *argv[]) {
 				const char *q = strchr(p + 1, '/');
 				if (q != NULL) {
 					qd.bailiwick = strdup(q + 1);
-					qd.rrtype = strndup(p + 1,
-							 (size_t)(q - p - 1));
+					qd.rrtype = strndup(p + 1, (size_t)
+							    (q - p - 1));
 				} else {
 					qd.rrtype = strdup(p + 1);
 				}
 				qd.thing = strndup(optarg,
-						      (size_t)(p - optarg));
+						   (size_t)(p - optarg));
 			} else {
 				qd.thing = strdup(optarg);
 			}
@@ -275,7 +302,7 @@ main(int argc, char *argv[]) {
 			const char *p = strchr(optarg, '/');
 			if (p != NULL) {
 				qd.thing = strndup(optarg,
-						      (size_t)(p - optarg));
+						   (size_t)(p - optarg));
 				qd.pfxlen = strdup(p + 1);
 			} else {
 				qd.thing = strdup(optarg);
@@ -294,8 +321,7 @@ main(int argc, char *argv[]) {
 				usage("-M must be positive");
 			break;
 		case 'u':
-			if ((psys = pick_system(optarg)) == NULL)
-				usage("-u must refer to a pdns system");
+			picked_system = strdup(optarg);
 			break;
 		case 'U':
 			donotverify = true;
@@ -308,12 +334,17 @@ main(int argc, char *argv[]) {
 			else if (strcasecmp(optarg, "text") == 0 ||
 				 strcasecmp(optarg, "dns") == 0)
 				presentation = pres_text;
+			else if (strcasecmp(optarg, "minimal") == 0)
+				presentation = pres_minimal;
 			else
-				usage("-p must specify json, text, or csv");
+				usage("-p must specify "
+				      "json, text, csv, or minimal");
+			DESTROY(presentation_name);
+			presentation_name = strdup(optarg);
 			break;
 		case 't':
 			if (qd.rrtype != NULL)
-				usage("can only specify rrtype one way");
+				usage("can only specify rrtype(s) once");
 			qd.rrtype = strdup(optarg);
 			break;
 		case 'b':
@@ -353,13 +384,15 @@ main(int argc, char *argv[]) {
 			break;
 		case 'j':
 			presentation = pres_json;
+			DESTROY(presentation_name);
+			presentation_name = strdup("json");
 			break;
 		case 'f':
 			switch (batching) {
 			case batch_none:
-				batching = batch_original;
+				batching = batch_terse;
 				break;
-			case batch_original:
+			case batch_terse:
 				batching = batch_verbose;
 				break;
 			case batch_verbose:
@@ -368,6 +401,23 @@ main(int argc, char *argv[]) {
 				usage("too many -f options");
 			}
 			break;
+		case 'T': {
+			char *copy, *walker, *token;
+			copy = walker = strdup(optarg);
+			while ((token = strsep(&walker, ",")) != NULL)
+				if (strcasecmp(token, "reverse") == 0)
+					transforms |= TRANS_REVERSE;
+				else if (strcasecmp(token, "datefix") == 0)
+					transforms |= TRANS_DATEFIX;
+				else if (strcasecmp(token, "chomp") == 0)
+					transforms |= TRANS_CHOMP;
+				else {
+					DESTROY(copy);
+					usage("unrecognized transform in -T");
+				}
+			DESTROY(copy);
+			break;
+		    }
 		case 'm':
 			multiple = true;
 			break;
@@ -415,14 +465,26 @@ main(int argc, char *argv[]) {
 			usage(msg);
 	}
 
-	/* recondition various options for HTML use. */
-	CURL *easy = curl_easy_init();
-	escape(easy, &qd.thing);
-	escape(easy, &qd.rrtype);
-	escape(easy, &qd.bailiwick);
-	escape(easy, &qd.pfxlen);
-	curl_easy_cleanup(easy);
-	easy = NULL;
+	if (asinfo_lookup) {
+#ifdef CRIPPLED_LIBC
+		usage("the -a option requires a modern functional C library.");
+#else
+		if (!asinfo_domain_exists(asinfo_domain)) {
+			fprintf(stderr,
+				"%s: ASINFO domain (%s) does not exist.\n",
+				program_name, asinfo_domain);
+			my_exit(1);
+		}
+#endif
+	}
+
+	if (presentation == pres_none) {
+		presentation = pres_text;
+		assert(presentation_name == NULL);
+		presentation_name = strdup("text");
+	}
+	if (presentation == pres_minimal)
+		minimal_deduper = deduper_new(minimal_modulus);
 
 	if ((msg = qparam_ready(&qp)) != NULL)
 		usage(msg);
@@ -446,29 +508,51 @@ main(int argc, char *argv[]) {
 	case pres_csv:
 		presenter = pverb->csv;
 		break;
+	case pres_minimal:
+		presenter = pverb->minimal;
+		break;
+	case pres_none:
+		/* FALLTHROUGH */
 	default:
 		abort();
 	}
+	assert(presentation != pres_none);
+	assert(presentation_name != NULL);
+	if (presenter == NULL) {
+		char *errmsg = NULL;
+		int x = asprintf(&errmsg,
+				 "that verb (%s) has no presenter for \"%s\"",
+				 pverb->name, presentation_name);
+		if (x < 0)
+			my_panic(true, "asprintf");
+		usage(errmsg);
+	}
 
 	/* get to final readiness; in particular, get psys set. */
 	if (sorting != no_sort)
 		sort_ready();
-	read_configs();
-	if (psys == NULL) {
-		psys = pick_system(DEFAULT_SYS);
-		if (psys == NULL)
-			usage("neither " DNSDBQ_SYSTEM
-			      " nor -u were specified,"
-			      " and there is no default.");
+	if (config_file == NULL)
+		config_file = select_config();
+	if (picked_system != NULL) {
+		psys_specified = true;
+		pick_system(picked_system, "-u option");
+		DESTROY(picked_system);
+	} else {
+		pick_system(DEFAULT_SYS, "default system");
+		psys_specified = true;
 	}
 
-	/* verify that some of the fields in our psys are set. */
-	assert(psys->base_url != NULL);
-	assert(psys->url != NULL);
-	assert(psys->status != NULL);
-	assert(psys->verb_ok != NULL);
-	assert(psys->ready != NULL);
-	assert(psys->destroy != NULL);
+	if (json_fd != -1) {
+#if WANT_PDNS_DNSDB
+		/* the json output files are in COF format, never SAF. */
+		if (strcmp(psys->name, "dnsdb2") == 0)
+			pick_system("dnsdb1", "downgrade for -J");
+#endif
+		NULL;
+	} else {
+		make_curl();
+		assert(psys_specified);
+	}
 
 	/* validate some interrelated options. */
 	if (multiple && batching == batch_none)
@@ -511,11 +595,7 @@ main(int argc, char *argv[]) {
 			usage("can't mix -t with -f");
 		if (info)
 			usage("can't mix -I with -f");
-		if ((msg = psys->ready()) != NULL)
-			usage(msg);
-		make_curl();
 		do_batch(stdin, &qp);
-		unmake_curl();
 	} else if (info) {
 		/* use the "info" verb. */
 		if (qd.mode != no_mode)
@@ -526,13 +606,9 @@ main(int argc, char *argv[]) {
 			usage("can't mix -b with -I");
 		if (qd.rrtype != NULL)
 			usage("can't mix -t with -I");
-		if (psys->info_req == NULL || psys->info_blob == NULL)
+		if (psys->info == NULL)
 			usage("there is no 'info' for this service");
-		if ((msg = psys->ready()) != NULL)
-			usage(msg);
-		make_curl();
-		psys->info_req();
-		unmake_curl();
+		psys->info();
 	} else {
 		/* do a LHS or RHS lookup of some kind. */
 		if (qd.mode == no_mode)
@@ -551,14 +627,15 @@ main(int argc, char *argv[]) {
 		if (qd.mode == ip_mode && qd.rrtype != NULL)
 			usage("can't mix -i with -t");
 
-		if ((msg = psys->ready()) != NULL)
-			usage(msg);
-		make_curl();
-		writer_t writer = writer_init(qp.output_limit);
+		writer_t writer = writer_init(qp.output_limit,
+					      ps_stdout, false);
 		(void) query_launcher(&qd, &qp, writer);
 		io_engine(0);
 		writer_fini(writer);
 		writer = NULL;
+	}
+
+	if (json_fd == -1) {
 		unmake_curl();
 	}
 
@@ -570,23 +647,16 @@ main(int argc, char *argv[]) {
 	my_exit(exit_code);
 }
 
-/* debug -- at the moment, dump to stderr.
- */
-void
-debug(bool want_header, const char *fmtstr, ...) {
-	va_list ap;
-
-	va_start(ap, fmtstr);
-	if (want_header)
-		fputs("debug: ", stderr);
-	vfprintf(stderr, fmtstr, ap);
-	va_end(ap);
-}
-
 /* my_exit -- close or destroy global objects, then exit.
  */
 __attribute__((noreturn)) void
 my_exit(int code) {
+	DESTROY(presentation_name);
+
+	/* deduper state if any can be trashed. */
+	if (minimal_deduper != NULL)
+		deduper_destroy(&minimal_deduper);
+
 	/* writers and readers which are still known, must be freed. */
 	unmake_writers();
 
@@ -594,12 +664,18 @@ my_exit(int code) {
 	unmake_curl();
 
 	/* globals which may have been initialized, are to be freed. */
+	DESTROY(config_file);
 	if (psys != NULL)
 		psys->destroy();
 
 	/* sort key specifications and computations, are to be freed. */
 	sort_destroy();
 
+#ifndef CRIPPLED_LIBC
+	/* asinfo logic has an internal DNS resolver context. */
+	asinfo_shutdown();
+#endif
+
 	/* terminate process. */
 	DEBUG(1, true, "about to call exit(%d)\n", code);
 	exit(code);
@@ -617,15 +693,6 @@ my_panic(bool want_perror, const char *s) {
 	my_exit(1);
 }
 
-/* or_else -- return one pointer or else the other.
- */
-const char *
-or_else(const char *p, const char *or_else) {
-	if (p != NULL)
-		return p;
-	return or_else;
-}
-
 /* Private. */
 
 /* help -- display a brief usage-help text; then exit.
@@ -636,31 +703,37 @@ static void
 help(void) {
 	verb_ct v;
 
-	printf("usage: %s [-cdfgGhIjmqSsUv468] [-p dns|json|csv]\n",
+	printf("usage: %s [-acdfGghIjmqSsUv468] [-p dns|json|csv|minimal]\n",
 	       program_name);
-	puts("\t[-k (first|last|duration|count|name|data)[,...]]\n"
-	     "\t[-l QUERY-LIMIT] [-L OUTPUT-LIMIT] [-A after] [-B before]\n"
-	     "\t[-u system] [-O offset] [-V verb] [-M max_count] {\n"
+	puts("\t[-u SYSTEM] [-V VERB] [-0 FUNCTION=INPUT]\n"
+	     "\t[-k (first|last|duration|count|name|type|data)[,...]]\n"
+	     "\t[-l QUERY-LIMIT] [-L OUTPUT-LIMIT]\n"
+	     "\t[-O OFFSET] [-M MAX_COUNT]\n"
+	     "\t[-A AFTER] [-B BEFORE]\n"
+	     "\t[-D ASINFO_DOMAIN] [-T (datefix|reverse|chomp)[,...] {\n"
 	     "\t\t-f |\n"
-	     "\t\t-J inputfile |\n"
-	     "\t\t[-t rrtype] [-b bailiwick] {\n"
-	     "\t\t\t-r OWNER[/TYPE[/BAILIWICK]] |\n"
-	     "\t\t\t-n NAME[/TYPE] |\n"
+	     "\t\t-J INPUTFILE |\n"
+	     "\t\t[-t RRTYPE[,...]] [-b BAILIWICK] {\n"
+	     "\t\t\t-r OWNER[/RRTYPE[,...][/BAILIWICK]] |\n"
+	     "\t\t\t-n NAME[/RRTYPE[,...]] |\n"
 	     "\t\t\t-i IP[/PFXLEN] |\n"
-	     "\t\t\t-N RAW-NAME-DATA[/TYPE]\n"
-	     "\t\t\t-R RAW-OWNER-DATA[/TYPE[/BAILIWICK]]\n"
+	     "\t\t\t-N RAW-NAME-DATA[/RRTYPE[,...]]\n"
+	     "\t\t\t-R RAW-OWNER-DATA[/RRTYPE[,...][/BAILIWICK]]\n"
 	     "\t\t}\n"
 	     "\t}");
-	puts("for -A and -B, use absolute format YYYY-MM-DD[ HH:MM:SS],\n"
-	     "\tor relative format %dw%dd%dh%dm%ds.\n"
+	printf("for -A and -B, use absolute format YYYY-MM-DD[ HH:MM:SS],\n"
+	     "\tor relative format %%dw%%dd%%dh%%dm%%ds.\n"
+	     "use -a to get ASNs associated with reported IP addresses\n"
 	     "use -c to get complete (strict) time matching for -A and -B.\n"
+	     "for -D, the default is \"%s\"\n"
 	     "use -d one or more times to ramp up the diagnostic output.\n"
+	     "for -0, the function must be \"countoff\"\n"
 	     "for -f, stdin must contain lines of the following forms:\n"
-	     "\trrset/name/NAME[/TYPE[/BAILIWICK]]\n"
-	     "\trrset/raw/HEX-PAIRS[/RRTYPE[/BAILIWICK]]\n"
-	     "\trdata/name/NAME[/TYPE]\n"
+	     "\trrset/name/NAME[/RRTYPE[,...][/BAILIWICK]]\n"
+	     "\trrset/raw/HEX-PAIRS[/RRTYPE[,...][/BAILIWICK]]\n"
+	     "\trdata/name/NAME[/RRTYPE[,...]]\n"
 	     "\trdata/ip/ADDR[,PFXLEN]\n"
-	     "\trdata/raw/HEX-PAIRS[/RRTYPE]\n"
+	     "\trdata/raw/HEX-PAIRS[/RRTYPE[,...]]\n"
 	     "\t(output format will depend on -p or -j, framed by '--'.)\n"
 	     "\t(with -ff, framing will be '++ $cmd', '-- $stat ($code)'.\n"
 	     "use -g to get graveled results (default is -G, rocks).\n"
@@ -677,17 +750,17 @@ help(void) {
 	     "use -s to sort in ascending order, "
 	     "or -S for descending order.\n"
 	     "\t-s/-S can be repeated before several -k arguments.\n"
+	     "for -T, transforms are datefix, reverse, and chomp.\n"
 	     "use -U to turn off SSL certificate verification.\n"
 	     "use -v to show the program version.\n"
 	     "use -4 to force connecting to the server via IPv4.\n"
 	     "use -6 to force connecting to the server via IPv6.\n"
-	     "use -8 to allow arbitrary 8-bit values in -r and -n arguments");
+	     "use -8 to allow 8-bit values in -r and -n arguments.\n",
+	     asinfo_domain);
 
 	puts("for -u, system must be one of:");
 #if WANT_PDNS_DNSDB
 	puts("\tdnsdb");
-#endif
-#if WANT_PDNS_DNSDB2
 	puts("\tdnsdb2");
 #endif
 #if WANT_PDNS_CIRCL
@@ -698,31 +771,10 @@ help(void) {
 		printf("\t%s\n", v->name);
 	puts("\nGetting Started:\n"
 	     "\tAdd your API key to ~/.dnsdb-query.conf like this:\n"
-	     "\t\tAPIKEY=\"YOURAPIKEYHERE\"");
+	     "\t\tDNSDB_API_KEY=\"YOURAPIKEYHERE\"");
 	printf("\nTry   man %s  for full documentation.\n", program_name);
 }
 
-/* pick_system -- return a named system descriptor, or NULL.
- */
-static pdns_system_ct
-pick_system(const char *name) {
-#if WANT_PDNS_DNSDB
-	if (strcmp(name, "dnsdb") == 0)
-		return pdns_dnsdb();
-#endif
-#if WANT_PDNS_DNSDB2
-	if (strcmp(name, "dnsdb2") == 0) {
-		encap = encap_saf;
-		return pdns_dnsdb2();
-	}
-#endif
-#if WANT_PDNS_CIRCL
-	if (strcmp(name, "circl") == 0)
-		return pdns_circl();
-#endif
-	return NULL;
-}
-
 /* qdesc_debug -- dump a qdesc.
  */
 static void
@@ -730,20 +782,24 @@ qdesc_debug(const char *where, qdesc_ct qdp) {
 	debug(true, "qdesc(%s)[", where);
 
 	const char *sep = "\040";
+	if (qdp->mode != no_mode) {
+		debug(false, "%smo %d", sep, (int)qdp->mode);
+		sep = ",\040";
+	}
 	if (qdp->thing != NULL) {
-		debug(true, "%sth '%s'", sep, qdp->thing);
+		debug(false, "%sth '%s'", sep, qdp->thing);
 		sep = ",\040";
 	}
 	if (qdp->rrtype != NULL) {
-		debug(true, "%srr '%s'\n", sep, qdp->rrtype);
+		debug(false, "%srr '%s'", sep, qdp->rrtype);
 		sep = ",\040";
 	}
 	if (qdp->bailiwick != NULL) {
-		debug(true, "%sbw '%s'\n", sep, qdp->bailiwick);
+		debug(false, "%sbw '%s'", sep, qdp->bailiwick);
 		sep = ",\040";
 	}
 	if (qdp->pfxlen != NULL) {
-		debug(true, "%spfx '%s'\n", sep, qdp->pfxlen);
+		debug(false, "%spfx '%s'", sep, qdp->pfxlen);
 		sep = ",\040";
 	}
 	debug(false, " ]\n");
@@ -832,9 +888,10 @@ static const char *
 qparam_ready(qparam_t qpp) {
 	if (qpp->output_limit == -1 && qpp->query_limit != -1 && !multiple)
 		qpp->output_limit = qpp->query_limit;
-	if (qpp->after != 0 && qpp->before != 0) {
+	if (qpp->complete && qpp->after != 0 && qpp->before != 0) {
 		if (qpp->after > qpp->before)
-			return "-A value must be before -B value (for now)";
+			return "-A value must be before -B value"
+			       " if using complete time matching";
 	}
 	if (qpp->complete && qpp->after == 0 && qpp->before == 0)
 		return "-c without -A or -B makes no sense.";
@@ -872,6 +929,7 @@ qparam_option(int opt, const char *arg, qparam_t qpp) {
 		if (!parse_long(arg, &qpp->output_limit) ||
 		    (qpp->output_limit <= 0))
 			return "-L must be positive";
+		qpp->explicit_output_limit = qpp->output_limit;
 		break;
 	case 'O':
 		if (!parse_long(optarg, &qpp->offset) ||
@@ -911,14 +969,14 @@ find_verb(const char *option) {
 
 	for (v = verbs; v->name != NULL; v++)
 		if (strcasecmp(option, v->name) == 0)
-			return (v);
-	return (NULL);
+			return v;
+	return NULL;
 }
 
-/* read_configs -- try to find a config file in static path, then parse it.
+/* select_config -- try to find a config file in static path.
  */
-static void
-read_configs(void) {
+static char *
+select_config(void) {
 	const char * const *conf;
 	char *cf = NULL;
 
@@ -930,121 +988,13 @@ read_configs(void) {
 		wordfree(&we);
 		if (access(cf, R_OK) == 0) {
 			DEBUG(1, true, "conf found: '%s'\n", cf);
-			break;
+			return cf;
 		}
 		DESTROY(cf);
 	}
-	if (cf != NULL) {
-		char *cmd, *line;
-		size_t n;
-		int x, l;
-		FILE *f;
-
-		/* in the "echo dnsdb server..." lines, the
-		 * first parameter is the pdns system to which to dispatch
-		 * the key and value (i.e. second the third parameters).
-		 */
-		x = asprintf(&cmd,
-			     ". %s;"
-			     "echo dnsdbq system $" DNSDBQ_SYSTEM ";"
-#if WANT_PDNS_DNSDB
-			     "echo dnsdb apikey $APIKEY;"
-			     "echo dnsdb server $DNSDB_SERVER;"
-#endif
-#if WANT_PDNS_DNSDB2
-			     "echo dnsdb2 apikey $APIKEY;"
-			     "echo dnsdb2 server $DNSDB_SERVER;"
-#endif
-#if WANT_PDNS_CIRCL
-			     "echo circl apikey $CIRCL_AUTH;"
-			     "echo circl server $CIRCL_SERVER;"
-#endif
-			     "exit", cf);
-		DESTROY(cf);
-		if (x < 0)
-			my_panic(true, "asprintf");
-		f = popen(cmd, "r");
-		if (f == NULL) {
-			fprintf(stderr, "%s: [%s]: %s",
-				program_name, cmd, strerror(errno));
-			DESTROY(cmd);
-			my_exit(1);
-		}
-		DEBUG(1, true, "conf cmd = '%s'\n", cmd);
-		DESTROY(cmd);
-		line = NULL;
-		n = 0;
-		l = 0;
-		while (getline(&line, &n, f) > 0) {
-			char *tok1, *tok2, *tok3;
-			char *saveptr = NULL;
-			const char *msg;
-
-			l++;
-			if (strchr(line, '\n') == NULL) {
-				fprintf(stderr,
-					"%s: conf line #%d: too long\n",
-					program_name, l);
-				my_exit(1);
-			}
-			tok1 = strtok_r(line, "\040\012", &saveptr);
-			tok2 = strtok_r(NULL, "\040\012", &saveptr);
-			tok3 = strtok_r(NULL, "\040\012", &saveptr);
-			if (tok1 == NULL || tok2 == NULL) {
-				fprintf(stderr,
-					"%s: conf line #%d: malformed\n",
-					program_name, l);
-				my_exit(1);
-			}
-			if (tok3 == NULL || *tok3 == '\0') {
-				/* variable wasn't set, ignore the line. */
-				continue;
-			}
-
-			/* some env/conf variables are dnsdbq-specific. */
-			if (strcmp(tok1, "dnsdbq") == 0) {
-				/* env/config psys does not override -u. */
-				if (psys == NULL &&
-				    strcmp(tok2, "system") == 0)
-				{
-					psys = pick_system(tok3);
-					if (psys == NULL) {
-						fprintf(stderr,
-							"%s: unknown %s %s\n",
-							program_name,
-							DNSDBQ_SYSTEM,
-							tok3);
-						my_exit(1);
-					}
-				}
-				continue;
-			}
-
-			/* this is the last point where psys can be null. */
-			if (psys == NULL) {
-				/* first match wins and is sticky. */
-				if ((psys = pick_system(tok1)) == NULL)
-					continue;
-				DEBUG(1, true, "picked system %s\n", tok1);
-			}
-
-			/* if this variable is for this system, consume it. */
-			if (strcmp(tok1, psys->name) == 0) {
-				DEBUG(1, true, "line #%d: sets %s|%s|%s\n",
-				      l, tok1, tok2,
-				      strcmp(tok2, "apikey") == 0
-					? "..." : tok3);
-				msg = psys->setval(tok2, tok3);
-				if (msg != NULL)
-					usage(msg);
-			}
-		}
-		DESTROY(line);
-		pclose(f);
-	}
+	return NULL;
 }
 
-
 /* do_batch -- implement "filter" mode, reading commands from a batch file.
  */
 static void
@@ -1057,7 +1007,7 @@ do_batch(FILE *f, qparam_ct qpp) {
 	/* if doing multiple parallel upstreams, start a writer. */
 	bool one_writer = multiple && batching != batch_verbose;
 	if (one_writer)
-		writer = writer_init(qp.output_limit);
+		writer = writer_init(qp.output_limit, ps_stdout, false);
 
 	while (getline(&command, &n, f) > 0) {
 		const char *msg;
@@ -1088,7 +1038,8 @@ do_batch(FILE *f, qparam_ct qpp) {
 
 		/* if not parallelizing, start a writer here instead. */
 		if (!one_writer)
-			writer = writer_init(qp.output_limit);
+			writer = writer_init(qp.output_limit,
+					     ps_stdout, false);
 
 		/* crack the batch line if possible. */
 		msg = batch_parse(command, &qd);
@@ -1096,16 +1047,22 @@ do_batch(FILE *f, qparam_ct qpp) {
 			fprintf(stderr, "%s: batch entry parse error: %s\n",
 				program_name, msg);
 		} else {
-			/* start one or two curl jobs based on this search. */
+			/* start one or more curl fetches based on this entry.
+			 */
 			query_t query = query_launcher(&qd, &qp, writer);
 
 			/* if merging, drain some jobs; else, drain all jobs.
 			 */
 			if (one_writer)
-				io_engine(MAX_JOBS);
+				io_engine(MAX_FETCHES);
 			else
 				io_engine(0);
-			if (query->status != NULL && batching != batch_verbose)
+
+			/* if one of our fetches already failed, say so now.
+			 */
+			if (query != NULL &&
+			    query->status != NULL &&
+			    batching != batch_verbose)
 			{
 				assert(query->message != NULL);
 				fprintf(stderr,
@@ -1123,7 +1080,7 @@ do_batch(FILE *f, qparam_ct qpp) {
 			switch (batching) {
 			case batch_none:
 				break;
-			case batch_original:
+			case batch_terse:
 				assert(writer->ps_buf == NULL &&
 				       writer->ps_len == 0);
 				writer->ps_buf = strdup("--\n");
@@ -1191,7 +1148,7 @@ batch_options(const char *optstr, qparam_t options, qparam_ct dflt) {
 		optind = 0;
 #else
 		/* 1 is the value that optind should be initialized to,
-		 * accorinng to IEEE Std 1003.1.
+		 * according to IEEE Std 1003.1.
 		 */
 		optind = 1;
 #if defined __FreeBSD__ || defined __OpenBSD__ || defined __NetBSD__ || \
@@ -1309,71 +1266,74 @@ batch_parse(char *line, qdesc_t qdp) {
 	return NULL;
 }
 
-/* makepath -- make a RESTful URI that describes these search parameters.
+/* makepath -- make a RESTful URI that describes these query parameters.
  *
  * Returns a string that must be free()d.
  */
 static char *
-makepath(mode_e mode, const char *name, const char *rrtype,
-	 const char *bailiwick, const char *pfxlen)
-{
-	char *command;
-	int x;
+makepath(qdesc_ct qdp) {
+	/* recondition various options for HTML use. */
+	char *thing = escape(qdp->thing);
+	char *rrtype = escape(qdp->rrtype);
+	char *bailiwick = escape(qdp->bailiwick);
+	char *pfxlen = escape(qdp->pfxlen);
 
-	switch (mode) {
+	char *path = NULL;
+	switch (qdp->mode) {
+		int x;
 	case rrset_mode:
 		if (rrtype != NULL && bailiwick != NULL)
-			x = asprintf(&command, "rrset/name/%s/%s/%s",
-				     name, rrtype, bailiwick);
+			x = asprintf(&path, "rrset/name/%s/%s/%s",
+				     thing, rrtype, bailiwick);
 		else if (rrtype != NULL)
-			x = asprintf(&command, "rrset/name/%s/%s",
-				     name, rrtype);
+			x = asprintf(&path, "rrset/name/%s/%s",
+				     thing, rrtype);
 		else if (bailiwick != NULL)
-			x = asprintf(&command, "rrset/name/%s/ANY/%s",
-				     name, bailiwick);
+			x = asprintf(&path, "rrset/name/%s/ANY/%s",
+				     thing, bailiwick);
 		else
-			x = asprintf(&command, "rrset/name/%s",
-				     name);
+			x = asprintf(&path, "rrset/name/%s",
+				     thing);
 		if (x < 0)
 			my_panic(true, "asprintf");
 		break;
 	case name_mode:
 		if (rrtype != NULL)
-			x = asprintf(&command, "rdata/name/%s/%s",
-				     name, rrtype);
+			x = asprintf(&path, "rdata/name/%s/%s",
+				     thing, rrtype);
 		else
-			x = asprintf(&command, "rdata/name/%s",
-				     name);
+			x = asprintf(&path, "rdata/name/%s",
+				     thing);
 		if (x < 0)
 			my_panic(true, "asprintf");
 		break;
 	case ip_mode:
 		if (pfxlen != NULL)
-			x = asprintf(&command, "rdata/ip/%s,%s",
-				     name, pfxlen);
+			x = asprintf(&path, "rdata/ip/%s,%s",
+				     thing, pfxlen);
 		else
-			x = asprintf(&command, "rdata/ip/%s",
-				     name);
+			x = asprintf(&path, "rdata/ip/%s",
+				     thing);
 		if (x < 0)
 			my_panic(true, "asprintf");
 		break;
 	case raw_rrset_mode:
 		if (rrtype != NULL)
-			x = asprintf(&command, "rrset/raw/%s/%s",
-				     name, rrtype);
+			x = asprintf(&path, "rrset/raw/%s/%s",
+				     thing, rrtype);
 		else
-			x = asprintf(&command, "rrset/raw/%s",
-				     name);
+			x = asprintf(&path, "rrset/raw/%s",
+				     thing);
 		if (x < 0)
 			my_panic(true, "asprintf");
 		break;
 	case raw_name_mode:
 		if (rrtype != NULL)
-			x = asprintf(&command, "rdata/raw/%s/%s",
-				     name, rrtype);
+			x = asprintf(&path, "rdata/raw/%s/%s",
+				     thing, rrtype);
 		else
-			x = asprintf(&command, "rdata/raw/%s",
-				     name);
+			x = asprintf(&path, "rdata/raw/%s",
+				     thing);
 		if (x < 0)
 			my_panic(true, "asprintf");
 		break;
@@ -1382,29 +1342,29 @@ makepath(mode_e mode, const char *name, const char *rrtype,
 	default:
 		abort();
 	}
-	return (command);
+
+	DESTROY(thing);
+	DESTROY(rrtype);
+	DESTROY(bailiwick);
+	DESTROY(pfxlen);
+
+	return path;
 }
 
 /* query_launcher -- fork off some curl jobs via launch() for this query.
+ *
+ * can write to STDERR and return NULL if a query cannot be launched.
  */
 static query_t
 query_launcher(qdesc_ct qdp, qparam_ct qpp, writer_t writer) {
 	struct pdns_fence fence = {};
 	query_t query = NULL;
+	const char *msg;
 
+	/* ready player one. */
 	CREATE(query, sizeof(struct query));
-	query->writer = writer;
-	writer = NULL;
-	query->params = *qpp;
-	query->next = query->writer->queries;
-	query->writer->queries = query;
-	query->command = makepath(qdp->mode, qdp->thing, qdp->rrtype,
-				  qdp->bailiwick, qdp->pfxlen);
 
-	/* figure out from time fencing which job(s) we'll be starting.
-	 *
-	 * the 4-tuple is: first_after, first_before, last_after, last_before
-	 */
+	/* define the fence. */
 	if (qpp->after != 0) {
 		if (qpp->complete) {
 			/* each db tuple must begin after the fence-start. */
@@ -1423,18 +1383,123 @@ query_launcher(qdesc_ct qdp, qparam_ct qpp, writer_t writer) {
 			fence.first_before = qpp->before;
 		}
 	}
-	launch(query, &fence);
+
+	/* branch on rrtype. */
+	if (qdp->rrtype == NULL) {
+		/* no rrtype string given, let makepath set it to "any". */
+		char *path = makepath(qdp);
+		launch_fetch(query, path, &fence);
+		DESTROY(path);
+	} else if ((msg = rrtype_correctness(qdp->rrtype)) != NULL) {
+		fprintf(stderr, "%s: rrtype incorrect: %s\n",
+			program_name, msg);
+		DESTROY(query);
+		return NULL;
+	} else {
+		/* rrtype string was given, parse comma separated list. */
+		char *rrtypes = strdup(qdp->rrtype);
+		char *saveptr = NULL;
+		int nfetches = 0;
+		for (char *rrtype = strtok_r(rrtypes, ",", &saveptr);
+		     rrtype != NULL;
+		     rrtype = strtok_r(NULL, ",", &saveptr))
+		{
+			struct qdesc qd = {
+				.mode = qdp->mode,
+				.thing = qdp->thing,
+				.rrtype = rrtype,
+				.bailiwick = qdp->bailiwick,
+				.pfxlen = qdp->pfxlen
+			};
+			char *path = makepath(&qd);
+			launch_fetch(query, path, &fence);
+			nfetches++;
+			DESTROY(path);
+		}
+		if (nfetches > 1)
+			query->multitype = true;
+		DESTROY(rrtypes);
+	}
+
+	/* finish query initialization, link it up, and return it. */
+	query->writer = writer;
+	writer = NULL;
+	query->params = *qpp;
+	query->next = query->writer->queries;
+	query->writer->queries = query;
+	query->descrip = makepath(qdp);
+	query->mode = qdp->mode;
 	return query;
 }
 
-/* launch -- actually launch a query job, given a command and time fences.
+/* rrtype_correctness -- return an error text if the rrtypes are senseless
+ */
+static const char *
+rrtype_correctness(const char *input) {
+	char **rrtypeset = calloc(MAX_FETCHES, sizeof(char *));
+	char *rrtypes = strdup(input);
+	char *saveptr = NULL;
+	const char *ret = NULL;
+	int nrrtypeset = 0;
+	bool some = false, any = false,
+		some_dnssec = false, any_dnssec = false;
+	for (char *rrtype = strtok_r(rrtypes, ",", &saveptr);
+	     rrtype != NULL;
+	     rrtype = strtok_r(NULL, ",", &saveptr))
+	{
+		for (char *p = rrtype; *p != '\0'; p++)
+			if (isupper(*p))
+				*p = (char) tolower(*p);
+		if (nrrtypeset == MAX_FETCHES) {
+			ret = "too many rrtypes specified";
+			goto done;
+		}
+		for (int i = 0; i < nrrtypeset; i++)
+			if (strcmp(rrtype, rrtypeset[i]) == 0) {
+				ret = "duplicate rrtype encountered";
+				goto done;
+			}
+		rrtypeset[nrrtypeset++] = rrtype;
+		if (strcmp(rrtype, "any") == 0)
+			any = true;
+		else if (strcmp(rrtype, "any-dnssec") == 0)
+			any_dnssec = true;
+		else if (strcmp(rrtype, "ds") == 0 ||
+			 strcmp(rrtype, "rrsig") == 0 ||
+			 strcmp(rrtype, "nsec") == 0 ||
+			 strcmp(rrtype, "dnskey") == 0 ||
+			 strcmp(rrtype, "cdnskey") == 0 ||
+			 strcmp(rrtype, "cds") == 0 ||
+			 strcmp(rrtype, "ta") == 0 ||
+			 strcmp(rrtype, "nsec3") == 0 ||
+			 strcmp(rrtype, "nsec3param") == 0 ||
+			 strcmp(rrtype, "dlv") == 0)
+			some_dnssec = true;
+		else
+			some = true;
+		if (any && some) {
+			ret = "ANY is redundant when mixed like this";
+			goto done;
+		}
+		if (any_dnssec && some_dnssec) {
+			ret = "ANY-DNSSEC is redundant when mixed like this";
+			goto done;
+		}
+	}
+ done:
+	DESTROY(rrtypes);
+	DESTROY(rrtypeset);
+	return ret;
+}
+
+/* launch_fetch -- actually launch a query job, given a path and time fences.
  */
 static void
-launch(query_t query, pdns_fence_ct fp) {
+launch_fetch(query_t query, const char *path, pdns_fence_ct fp) {
 	qparam_ct qpp = &query->params;
-	char *url, sep;
+	char *url;
 
-	url = psys->url(query->command, &sep, qpp, fp);
+	url = psys->url(path, NULL, qpp, fp, false);
 	if (url == NULL)
 		my_exit(1);
 
@@ -1453,7 +1518,7 @@ ruminate_json(int json_fd, qparam_ct qpp) {
 	writer_t writer;
 	ssize_t len;
 
-	writer = writer_init(qpp->output_limit);
+	writer = writer_init(qpp->output_limit, NULL, false);
 	CREATE(query, sizeof(struct query));
 	query->writer = writer;
 	query->params = *qpp;
@@ -1462,7 +1527,7 @@ ruminate_json(int json_fd, qparam_ct qpp) {
 	query->fetches = fetch;
 	writer->queries = query;
 	CREATE(buf, ideal_buffer);
-	while ((len = read(json_fd, buf, sizeof buf)) > 0) {
+	while ((len = read(json_fd, buf, ideal_buffer)) > 0) {
 		writer_func(buf, 1, (size_t)len, query->fetches);
 	}
 	DESTROY(buf);
