@@ -35,12 +35,12 @@ static void io_drain(void);
 static void fetch_reap(fetch_t);
 static void fetch_done(fetch_t);
 static void fetch_unlink(fetch_t);
-static void query_done(query_t);
+static void last_fetch(fetch_t);
 
 static writer_t writers = NULL;
 static CURLM *multi = NULL;
 static bool curl_cleanup_needed = false;
-static query_t paused[MAX_JOBS];
+static query_t paused[MAX_FETCHES];
 static int npaused = 0;
 
 const char saf_begin[] = "begin";
@@ -83,7 +83,7 @@ unmake_curl(void) {
 
 /* fetch -- given a url, tell libcurl to go fetch it.
  */
-void
+fetch_t
 create_fetch(query_t query, char *url) {
 	fetch_t fetch = NULL;
 	CURLMcode res;
@@ -115,7 +115,7 @@ create_fetch(query_t query, char *url) {
 	if (psys->auth != NULL)
 	    psys->auth(fetch);
 
-	if (encap == encap_saf)
+	if (psys->encap == encap_saf)
 		fetch->hdrs = curl_slist_append(fetch->hdrs, jsonl_header);
 	else
 		fetch->hdrs = curl_slist_append(fetch->hdrs, json_header);
@@ -123,9 +123,12 @@ create_fetch(query_t query, char *url) {
 	curl_easy_setopt(fetch->easy, CURLOPT_WRITEFUNCTION, writer_func);
 	curl_easy_setopt(fetch->easy, CURLOPT_WRITEDATA, fetch);
 	curl_easy_setopt(fetch->easy, CURLOPT_PRIVATE, fetch);
+#ifdef CURL_AT_LEAST_VERSION
+/* If CURL_AT_LEAST_VERSION is not defined then the curl is probably too old */
 #if CURL_AT_LEAST_VERSION(7,42,0)
 	/* do not allow curl to swallow /./ and /../ in our URLs */
 	curl_easy_setopt(fetch->easy, CURLOPT_PATH_AS_IS, 1L);
+#endif
 #endif /* CURL_AT_LEAST_VERSION */
 	if (debug_level >= 3)
 		curl_easy_setopt(fetch->easy, CURLOPT_VERBOSE, 1L);
@@ -140,6 +143,7 @@ create_fetch(query_t query, char *url) {
 			program_name, curl_multi_strerror(res));
 		my_exit(1);
 	}
+	return fetch;
 }
 
 /* fetch_reap -- reap one fetch.
@@ -168,7 +172,7 @@ fetch_done(fetch_t fetch) {
 
 	/* if this was the last fetch on some query, signal. */
 	if (query->fetches == fetch && fetch->next == NULL)
-		query_done(query);
+		last_fetch(fetch);
 }
 
 /* fetch_unlink -- disconnect a fetch from its writer.
@@ -191,11 +195,13 @@ fetch_unlink(fetch_t fetch) {
 /* writer_init -- instantiate a writer, which may involve forking a "sort".
  */
 writer_t
-writer_init(long output_limit) {
+writer_init(long output_limit, ps_user_t ps_user, bool meta_query) {
 	writer_t writer = NULL;
 
 	CREATE(writer, sizeof(struct writer));
 	writer->output_limit = output_limit;
+	writer->ps_user = ps_user;
+	writer->meta_query = meta_query;
 
 	if (sorting != no_sort) {
 		/* sorting involves a subprocess (POSIX sort(1) command),
@@ -222,7 +228,12 @@ writer_init(long output_limit) {
 
 	writer->next = writers;
 	writers = writer;
-	return (writer);
+	return writer;
+}
+
+void
+ps_stdout(writer_t writer) {
+	fwrite(writer->ps_buf, 1, writer->ps_len, stdout);
 }
 
 /* query_status -- install a status code and description in a query.
@@ -230,7 +241,12 @@ writer_init(long output_limit) {
 void
 query_status(query_t query, const char *status, const char *message) {
 	assert((query->status == NULL) == (query->message == NULL));
-	assert(query->status == NULL);
+	if (query->multitype && query->status != NULL) {
+		DESTROY(query->status);
+		DESTROY(query->message);
+	} else {
+		assert(query->status == NULL);
+	}
 	query->status = strdup(status);
 	query->message = strdup(message);
 }
@@ -263,17 +279,17 @@ writer_func(char *ptr, size_t size, size_t nmemb, void *blob) {
 				/* grab the token. */
 				writer->active = query;
 				DEBUG(2, true, "active (%d) %s\n",
-				      npaused, query->command);
+				      npaused, query->descrip);
 			} else if (writer->active != query) {
 				/* pause the query. */
 				paused[npaused++] = query;
 				DEBUG(2, true, "pause (%d) %s\n",
-				      npaused, query->command);
+				      npaused, query->descrip);
 				return CURL_WRITEFUNC_PAUSE;
 			}
 		}
 		if (!query->hdr_sent) {
-			printf("++ %s\n", query->command);
+			printf("++ %s\n", query->descrip);
 			query->hdr_sent = true;
 		}
 	}
@@ -298,31 +314,26 @@ writer_func(char *ptr, size_t size, size_t nmemb, void *blob) {
 			if (eol != NULL)
 				*eol = '\0';
 
-			/* only report the first response status (vs. -m). */
-			if (query->status == NULL) {
+			/* only remember the first response status. */
+			if (query->status == NULL)
 				query_status(query,
 					     psys->status(fetch),
 					     message);
-				if (!quiet) {
-					char *url;
+			if (!quiet) {
+				char *url;
 
-					curl_easy_getinfo(fetch->easy,
-							CURLINFO_EFFECTIVE_URL,
-							  &url);
-					fprintf(stderr,
-						"%s: warning: "
-						"libcurl %ld [%s]\n",
-						program_name, fetch->rcode,
-						url);
-				}
+				curl_easy_getinfo(fetch->easy,
+						  CURLINFO_EFFECTIVE_URL,
+						  &url);
+				fprintf(stderr,
+					"%s: warning: libcurl %ld [%s] %s\n",
+					program_name, fetch->rcode,
+					url, message);
 			}
-			if (!quiet)
-				fprintf(stderr, "%s: warning: libcurl: [%s]\n",
-					program_name, message);
 			DESTROY(message);
 			fetch->buf[0] = '\0';
 			fetch->len = 0;
-			return (bytes);
+			return bytes;
 		}
 	}
 
@@ -338,26 +349,22 @@ writer_func(char *ptr, size_t size, size_t nmemb, void *blob) {
 			      qp->output_limit);
 			/* cause CURLE_WRITE_ERROR for this transfer. */
 			bytes = 0;
-			if (encap == encap_saf)
-				query->saf_cond = sc_we_limited;
+			if (psys->encap == encap_saf)
+				fetch->saf_cond = sc_we_limited;
 			/* inform io_engine() that the abort is intentional. */
 			fetch->stopped = true;
-		} else if (writer->info) {
-			/* concatenate this fragment (with \n) to info_buf. */
-			char *temp = NULL;
-
-			asprintf(&temp, "%s%*.*s\n",
-				 or_else(writer->ps_buf, ""),
-				 (int)pre_len, (int)pre_len, fetch->buf);
-			DESTROY(writer->ps_buf);
-			writer->ps_buf = temp;
+		} else if (writer->meta_query) {
+			/* concatenate this fragment (incl \n) to ps_buf. */
+			writer->ps_buf = realloc(writer->ps_buf,
+						 writer->ps_len + pre_len + 1);
+			memcpy(writer->ps_buf + writer->ps_len,
+			       fetch->buf, pre_len + 1);
 			writer->ps_len += pre_len + 1;
 		} else {
-			query->writer->count +=
-				data_blob(query, fetch->buf, pre_len);
+			query->writer->count += data_blob(fetch, pre_len);
 
-			if (encap == encap_saf)
-				switch (query->saf_cond) {
+			if (psys->encap == encap_saf)
+				switch (fetch->saf_cond) {
 				case sc_init:
 				case sc_begin:
 				case sc_ongoing:
@@ -377,25 +384,31 @@ writer_func(char *ptr, size_t size, size_t nmemb, void *blob) {
 		fetch->len = post_len;
 	}
 
-	return (bytes);
+	return bytes;
 }
 
-/* query_done -- do something with leftover buffer data when a query ends.
+/* last_fetch -- do something with leftover buffer data when a query ends.
  */
 static void
-query_done(query_t query) {
-	DEBUG(2, true, "query_done(%s)\n", query->command);
+last_fetch(fetch_t fetch) {
+	query_t query = fetch->query;
+
+	assert(query->fetches == fetch && fetch->next == NULL);
+	DEBUG(2, true, "query_done(%s), meta=%d\n",
+	      query->descrip, query->writer->meta_query);
+	if (query->writer->meta_query)
+		return;
 
 	if (batching == batch_none && !quiet) {
-		const char *msg = or_else(query->saf_msg, "");
+		const char *msg = or_else(fetch->saf_msg, "");
 
-		if (query->saf_cond == sc_limited)
-			fprintf(stderr, "Query limited: %s\n", msg);
-		else if (query->saf_cond == sc_failed)
-			fprintf(stderr, "Query failed: %s\n", msg);
-		else if (query->saf_cond == sc_missing)
-			fprintf(stderr, "Query response_missing: %s\n", msg);
-		else if (query->status != NULL)
+		if (fetch->saf_cond == sc_limited)
+			fprintf(stderr, "Database limit: %s\n", msg);
+		else if (fetch->saf_cond == sc_failed)
+			fprintf(stderr, "Database result: %s\n", msg);
+		else if (fetch->saf_cond == sc_missing)
+			fprintf(stderr, "Query response missing: %s\n", msg);
+		else if (query->status != NULL && !query->multitype)
 			fprintf(stderr, "Query status: %s (%s)\n",
 				query->status, query->message);
 	} else if (batching == batch_verbose) {
@@ -411,10 +424,10 @@ query_done(query_t query) {
 			asprintf(&writer->ps_buf, "-- %s (%s)\n",
 				 or_else(query->status, status_noerror),
 				 or_else(query->message,
-					 or_else(query->saf_msg, "no error")));
+					 or_else(fetch->saf_msg, "no error")));
 		if (npaused > 0) {
 			query_t unpause;
-			fetch_t fetch;
+			fetch_t ufetch;
 			int i;
 
 			/* unpause the next query's fetches. */
@@ -422,12 +435,12 @@ query_done(query_t query) {
 			npaused--;
 			for (i = 0; i < npaused; i++)
 				paused[i] = paused[i + 1];
-			for (fetch = unpause->fetches;
-			     fetch != NULL;
-			     fetch = fetch->next) {
+			for (ufetch = unpause->fetches;
+			     ufetch != NULL;
+			     ufetch = ufetch->next) {
 				DEBUG(2, true, "unpause (%d) %s\n",
-				      npaused, unpause->command);
-				curl_easy_pause(fetch->easy, CURLPAUSE_CONT);
+				      npaused, unpause->descrip);
+				curl_easy_pause(ufetch->easy, CURLPAUSE_CONT);
 			}
 		}
 	}
@@ -472,6 +485,9 @@ writer_fini(writer_t writer) {
 				fetch->len = 0;
 			}
 
+			/* release any fetch-specific data. */
+			DESTROY(fetch->saf_msg);
+
 			/* tear down any curl infrastructure on the fetch. */
 			fetch_reap(fetch);
 			fetch = NULL;
@@ -480,7 +496,7 @@ writer_fini(writer_t writer) {
 		assert((query->status != NULL) == (query->message != NULL));
 		DESTROY(query->status);
 		DESTROY(query->message);
-		DESTROY(query->command);
+		DESTROY(query->descrip);
 		DESTROY(query);
 		writer->queries = query_next;
 	}
@@ -530,7 +546,7 @@ writer_fini(writer_t writer) {
 				 (int)(nl - linep),
 				 (int)(nl - linep),
 				 linep);
-			/* skip sort keys (first, last, count, name, data). */
+			/* skip sort key: first */
 			if ((linep = strchr(linep, ' ')) == NULL) {
 				fprintf(stderr,
 					"%s: warning: no SP found in '%s'\n",
@@ -538,6 +554,7 @@ writer_fini(writer_t writer) {
 				continue;
 			}
 			linep += strspn(linep, " ");
+			/* skip sort key: last */
 			if ((linep = strchr(linep, ' ')) == NULL) {
 				fprintf(stderr,
 					"%s: warning: no second SP in '%s'\n",
@@ -545,6 +562,7 @@ writer_fini(writer_t writer) {
 				continue;
 			}
 			linep += strspn(linep, " ");
+			/* skip sort key: duration */
 			if ((linep = strchr(linep, ' ')) == NULL) {
 				fprintf(stderr,
 					"%s: warning: no third SP in '%s'\n",
@@ -552,6 +570,7 @@ writer_fini(writer_t writer) {
 				continue;
 			}
 			linep += strspn(linep, " ");
+			/* skip sort key: count */
 			if ((linep = strchr(linep, ' ')) == NULL) {
 				fprintf(stderr,
 					"%s: warning: no fourth SP in '%s'\n",
@@ -559,6 +578,7 @@ writer_fini(writer_t writer) {
 				continue;
 			}
 			linep += strspn(linep, " ");
+			/* skip sort key: rrname */
 			if ((linep = strchr(linep, ' ')) == NULL) {
 				fprintf(stderr,
 					"%s: warning: no fifth SP in '%s'\n",
@@ -566,6 +586,7 @@ writer_fini(writer_t writer) {
 				continue;
 			}
 			linep += strspn(linep, " ");
+			/* skip sort key: rrtype */
 			if ((linep = strchr(linep, ' ')) == NULL) {
 				fprintf(stderr,
 					"%s: warning: no sixth SP in '%s'\n",
@@ -573,6 +594,30 @@ writer_fini(writer_t writer) {
 				continue;
 			}
 			linep += strspn(linep, " ");
+			/* skip sort key: rdata */
+			if ((linep = strchr(linep, ' ')) == NULL) {
+				fprintf(stderr,
+					"%s: warning: no seventh SP in '%s'\n",
+					program_name, line);
+				continue;
+			}
+			linep += strspn(linep, " ");
+			/* recover the mode */
+			mode_e mode = (mode_e) (int) strtol(linep, NULL, 10);
+			if (mode == no_mode) {
+				fprintf(stderr,
+					"%s: invalid mode from sort in '%s'\n",
+					program_name, line);
+				continue;
+			}
+			if ((linep = strchr(linep, ' ')) == NULL) {
+				fprintf(stderr,
+					"%s: warning: no eighth SP in '%s'\n",
+					program_name, line);
+				continue;
+			}
+			linep += strspn(linep, " ");
+			/* recover the json */
 			DEBUG(2, true, "sort2: '%*.*s'\n",
 				 (int)(nl - linep),
 				 (int)(nl - linep),
@@ -585,7 +630,7 @@ writer_fini(writer_t writer) {
 					program_name, msg);
 				continue;
 			}
-			(*presenter)(&tup, linep, len, writer);
+			(*presenter)(&tup, mode, writer);
 			tuple_unmake(&tup);
 			count++;
 		}
@@ -606,10 +651,8 @@ writer_fini(writer_t writer) {
 
 	/* burp out the stored postscript, if any, and destroy it. */
 	if (writer->ps_len > 0) {
-		if (writer->info)
-			psys->info_blob(writer->ps_buf, writer->ps_len);
-		else
-			fwrite(writer->ps_buf, 1, writer->ps_len, stdout);
+		assert(writer->ps_user != NULL);
+		writer->ps_user(writer);
 		DESTROY(writer->ps_buf);
 		writer->ps_len = 0;
 	}
@@ -686,12 +729,28 @@ io_drain(void) {
 						  &fetch->rcode);
 
 			DEBUG(2, true, "io_drain(%s) DONE rcode=%d\n",
-			      query->command, fetch->rcode);
-			if (encap == encap_saf)
+			      query->descrip, fetch->rcode);
+			if (psys->encap == encap_saf) {
+				if (fetch->saf_cond == sc_begin ||
+				    fetch->saf_cond == sc_ongoing)
+				{
+					/* stream ended without a terminating
+					 * SAF value, so override stale value
+					 * we received before the problem.
+					 */
+					fetch->saf_cond = sc_missing;
+					fetch->saf_msg = strdup(
+						"Data transfer failed "
+						"-- No SAF terminator "
+						"at end of stream");
+					query_status(query,
+						     status_error,
+						     fetch->saf_msg);
+				}
 				DEBUG(2, true, "... saf_cond %d saf_msg %s\n",
-				      query->saf_cond,
-				      or_else(query->saf_msg, ""));
-
+				      fetch->saf_cond,
+				      or_else(fetch->saf_msg, ""));
+			}
 			if (cm->data.result == CURLE_COULDNT_RESOLVE_HOST) {
 				fprintf(stderr,
 					"%s: warning: libcurl failed since "
@@ -716,8 +775,9 @@ io_drain(void) {
 			}
 
 			/* record emptiness as status if nothing else. */
-			if (encap == encap_saf &&
+			if (psys->encap == encap_saf &&
 			    query->writer != NULL &&
+			    !query->writer->meta_query &&
 			    query->writer->count == 0 &&
 			    query->status == NULL)
 			{
@@ -734,22 +794,22 @@ io_drain(void) {
 	}
 }
 
-/* escape -- HTML-encode a string, in place.
+/* escape -- HTML-encode a string, returns a string which must be free()'d.
  */
-void
-escape(CURL *easy, char **str) {
-	char *escaped;
+char *
+escape(const char *str) {
+	char *escaped, *ret;
 
-	if (*str == NULL)
-		return;
-	escaped = curl_easy_escape(easy, *str, (int)strlen(*str));
+	if (str == NULL)
+		return NULL;
+	escaped = curl_escape(str, (int)strlen(str));
 	if (escaped == NULL) {
 		fprintf(stderr, "%s: curl_escape(%s) failed\n",
-			program_name, *str);
+			program_name, str);
 		my_exit(1);
 	}
-	DESTROY(*str);
-	*str = strdup(escaped);
+	ret = strdup(escaped);
 	curl_free(escaped);
 	escaped = NULL;
+	return ret;
 }
